/*
 * @(#)Tetris.c
 *
 * Copyright 1993 - 2010  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Methods file for Tetris */

#include "file.h"
#include "rngs.h"
#include "sound.h"
#include "TetrisP.h"

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wtetris.ini"
#endif

#define SECTION "setup"
#else

#if defined( USE_SOUND ) && defined( USE_NAS )
Display *dsp;
#endif

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void resizeGame(TetrisWidget w);
static void sizeGame(TetrisWidget w);
static Boolean setValuesGame(Widget current, Widget request, Widget renew);
static void destroyGame(Widget old);
static void quitGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void initializeGame(Widget request, Widget renew);
static void exposeGame(Widget renew, XEvent *event, Region region);
static void mouseGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void hideGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void clearGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void enterGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void leaveGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void reflectGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void rotateGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void moveGameLeft(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void moveGameRight(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void dropGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void moveGameInput(TetrisWidget w, int direction);
static void incrementGameLevel(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameLevel(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameUnits(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameUnits(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameColumns(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameColumns(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void fixedfillGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void suddenAppearGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void diagonalGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void mixedGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void bonusGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void cwGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void plainPieceGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextGame(TetrisWidget w,
	XEvent *event, char **args, int nArgs);
static void soundGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void pauseGame(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill0Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill1Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill2Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill3Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill4Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill5Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill6Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill7Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill8Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill9Game(TetrisWidget w, XEvent *event, char **args, int nArgs);
static void fill10Game(TetrisWidget w, XEvent *event, char **args, int nArgs);

static char defaultTranslationsGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <Btn4Down>: Mouse() Reflect()\n\
 <KeyPress>i: Reflect()\n\
 <KeyPress>e: Reflect()\n\
 <KeyPress>F11: Reflect()\n\
 <KeyPress>KP_Divide: Reflect()\n\
 <KeyPress>R5: Reflect()\n\
 <Btn2Down>: Mouse() Rotate()\n\
 <KeyPress>k: Rotate()\n\
 <KeyPress>d: Rotate()\n\
 <KeyPress>Up: Rotate()\n\
 <KeyPress>osfUp: Rotate()\n\
 <KeyPress>KP_Up: Rotate()\n\
 <KeyPress>R8: Rotate()\n\
 <KeyPress>KP_8: Rotate()\n\
 <KeyPress>F12: Rotate()\n\
 <KeyPress>Begin: Rotate()\n\
 <KeyPress>KP_5: Rotate()\n\
 <KeyPress>R11: Rotate()\n\
 <Btn1Down>: Mouse() MoveLeft()\n\
 <KeyPress>j: MoveLeft()\n\
 <KeyPress>s: MoveLeft()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <Btn3Down>: Mouse() MoveRight()\n\
 <KeyPress>l: MoveRight()\n\
 <KeyPress>f: MoveRight()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>Return: Drop()\n\
 <KeyPress>KP_Enter: Drop()\n\
 <KeyPress>Down: Drop()\n\
 <KeyPress>osfDown: Drop()\n\
 <KeyPress>KP_Down: Drop()\n\
 <KeyPress>KP_2: Drop()\n\
 <KeyPress>R14: Drop()\n\
 <KeyPress>KP_Space: Drop()\n\
 <KeyPress>space: Drop()\n\
 <Btn4Down>: Mouse() Rotate()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>>: IncrementLevel()\n\
 <KeyPress>.: IncrementLevel()\n\
 <KeyPress><: DecrementLevel()\n\
 <KeyPress>+: IncrementUnits()\n\
 <KeyPress>=: IncrementUnits()\n\
 <KeyPress>-: DecrementUnits()\n\
 <KeyPress>_: DecrementUnits()\n\
 <KeyPress>]: IncrementColumns()\n\
 <KeyPress>}: IncrementColumns()\n\
 <KeyPress>[: DecrementColumns()\n\
 <KeyPress>{: DecrementColumns()\n\
 <KeyPress>x: FixedFill()\n\
 <KeyPress>r: SuddenAppear()\n\
 <KeyPress>g: Diagonal()\n\
 <KeyPress>m: Mixed()\n\
 <KeyPress>b: Bonus()\n\
 <KeyPress>w: Cw()\n\
 <KeyPress>p: PlainPiece()\n\
 <KeyPress>n: ShowNext()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>|: Pause()\n\
 <KeyPress>`: Fill0()\n\
 <KeyPress>1: Fill1()\n\
 <KeyPress>2: Fill2()\n\
 <KeyPress>3: Fill3()\n\
 <KeyPress>4: Fill4()\n\
 <KeyPress>5: Fill5()\n\
 <KeyPress>6: Fill6()\n\
 <KeyPress>7: Fill7()\n\
 <KeyPress>8: Fill8()\n\
 <KeyPress>9: Fill9()\n\
 <KeyPress>0: Fill10()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

/*
 <KeyPress>,: DecrementLevel()\n\
#if defined(UP_REFLECT)
 <KeyPress>KP_Up: Reflect()\n\
 <KeyPress>R8: Reflect()\n\
 <KeyPress>KP_8: Reflect()\n\
#endif
*/

static XtActionsRec actionsListGame[] =
{
	{(char *) "Quit", (XtActionProc) quitGame},
	{(char *) "Hide", (XtActionProc) hideGame},
	{(char *) "Mouse", (XtActionProc) mouseGame},
	{(char *) "Reflect", (XtActionProc) reflectGame},
	{(char *) "Rotate", (XtActionProc) rotateGame},
	{(char *) "MoveLeft", (XtActionProc) moveGameLeft},
	{(char *) "MoveRight", (XtActionProc) moveGameRight},
	{(char *) "Drop", (XtActionProc) dropGame},
	{(char *) "Clear", (XtActionProc) clearGame},
	{(char *) "IncrementLevel", (XtActionProc) incrementGameLevel},
	{(char *) "DecrementLevel", (XtActionProc) decrementGameLevel},
	{(char *) "IncrementUnits", (XtActionProc) incrementGameUnits},
	{(char *) "DecrementUnits", (XtActionProc) decrementGameUnits},
	{(char *) "IncrementColumns", (XtActionProc) incrementGameColumns},
	{(char *) "DecrementColumns", (XtActionProc) decrementGameColumns},
	{(char *) "FixedFill", (XtActionProc) fixedfillGame},
	{(char *) "SuddenAppear", (XtActionProc) suddenAppearGame},
	{(char *) "Diagonal", (XtActionProc) diagonalGame},
	{(char *) "Mixed", (XtActionProc) mixedGame},
	{(char *) "Bonus", (XtActionProc) bonusGame},
	{(char *) "Cw", (XtActionProc) cwGame},
	{(char *) "PlainPiece", (XtActionProc) plainPieceGame},
	{(char *) "ShowNext", (XtActionProc) showNextGame},
	{(char *) "Sound", (XtActionProc) soundGame},
	{(char *) "Pause", (XtActionProc) pauseGame},
	{(char *) "Fill0", (XtActionProc) fill0Game},
	{(char *) "Fill1", (XtActionProc) fill1Game},
	{(char *) "Fill2", (XtActionProc) fill2Game},
	{(char *) "Fill3", (XtActionProc) fill3Game},
	{(char *) "Fill4", (XtActionProc) fill4Game},
	{(char *) "Fill5", (XtActionProc) fill5Game},
	{(char *) "Fill6", (XtActionProc) fill6Game},
	{(char *) "Fill7", (XtActionProc) fill7Game},
	{(char *) "Fill8", (XtActionProc) fill8Game},
	{(char *) "Fill9", (XtActionProc) fill9Game},
	{(char *) "Fill10", (XtActionProc) fill10Game},
	{(char *) "Enter", (XtActionProc) enterGame},
	{(char *) "Leave", (XtActionProc) leaveGame}
};

static XtResource resourcesGame[] =
{
	/* 30*10+4,24*10+4 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(TetrisWidget, core.width),
	 XtRString, (caddr_t) "306"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(TetrisWidget, core.height),
	 XtRString, (caddr_t) "606"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(TetrisWidget, tetris.background),
	 XtRString, (caddr_t) "#AEB2C3" /*XtDefaultBackground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.delay),
	 XtRString, (caddr_t) "1000"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbumpSound, XtCBumpSound, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.bumpSound),
	 XtRString, (caddr_t) BUMPSOUND},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNlevel, XtCLevel, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.level),
	 XtRString, (caddr_t) "3"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.units),
	 XtRString, (caddr_t) "4"},
	{XtNfill, XtCFill, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.fill),
	 XtRString, (caddr_t) "0"},
	{XtNfixedFill, XtCFixedFill, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.fixedFill),
	 XtRString, (caddr_t) "TRUE"},
	{XtNsuddenAppear, XtCSuddenAppear, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.suddenAppear),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.bonus),
	 XtRString, (caddr_t) "TRUE"},
	{XtNcw, XtCCw, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.cw),
	 XtRString, (caddr_t) "TRUE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNscore, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.score),
	 XtRString, (caddr_t) "0"},
	{XtNcompletedRows, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.completedRows),
	 XtRString, (caddr_t) "0"},
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(TetrisWidget, tetris.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNact, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TetrisWidget, tetris.act),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNpixmapSize, XtCPixmapSize, XtRInt, sizeof (int),
	 XtOffset(TetrisWidget, tetris.pixmapSize),
	 XtRString, (caddr_t) "64"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(TetrisWidget, tetris.select),
	 XtRCallback, (caddr_t) NULL}
};

TetrisClassRec tetrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Tetris",	/* class name */
		sizeof (TetrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListGame,	/* actions */
		XtNumber(actionsListGame),	/* num actions */
		resourcesGame,	/* resources */
		XtNumber(resourcesGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyGame,	/* destroy */
		(XtWidgetProc) resizeGame,	/* resize */
		(XtExposeProc) exposeGame,	/* expose */
		(XtSetValuesFunc) setValuesGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass tetrisWidgetClass = (WidgetClass) & tetrisClassRec;

static void
setGame(TetrisWidget w, int reason)
{
	tetrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

static void
setGameNext(TetrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	tetrisCallbackStruct cb;

	cb.reason = reason;
	cb.randomNumber = randomNumber;
	cb.randomRotation = randomRotation;
	cb.randomReflection = randomReflection;
	cb.units = units;
	cb.diagonal = diagonal;
	cb.mixed = mixed;
	cb.newBonus = newBonus;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static const int delay[MAX_LEVELS] = {
	100, 92, 84, 75, 68, 60, 52, 45, 39, 32,
	26, 21, 16, 12, 8, 5, 3, 2, 1, 0
};

static void
checkThings(TetrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;


	if (w->tetris.level < MIN_LEVELS) {
		w->tetris.level = MIN_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at least ",
			MIN_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->tetris.level > MAX_LEVELS) {
		w->tetris.level = MAX_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at most ",
			MAX_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tetris.units < MIN_UNITS) {
		w->tetris.units = MIN_UNITS;
		w->tetris.next.units = w->tetris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->tetris.units > MAX_UNITS) {
		w->tetris.units = MAX_UNITS;
		w->tetris.next.units = w->tetris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tetris.columns < MIN_COLUMNS) {
		w->tetris.columns = MIN_COLUMNS;
		intCat(&buf1, "Number of columns out of bounds, use at least ",
			MIN_COLUMNS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.columns);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (!LEGAL(w->tetris.units, w->tetris.diagonal,
			w->tetris.mixed, FALSE)) {
		intCat(&buf1, "Illegal configuration of units ",
			w->tetris.units);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tetris.diagonal);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->tetris.mixed);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to units ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_DIAGONAL);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_MIXED);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->tetris.units = DEFAULT_UNITS;
		w->tetris.diagonal = FALSE;
		w->tetris.mixed = FALSE;
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.mixed = w->tetris.mixed;
	}
	if (w->tetris.fill < 0) {
		w->tetris.fill = 0;
		DISPLAY_WARNING("Number for fill can not be negative, defaulting to 0");
	} else if (w->tetris.fill > MAX_FILL) {
		w->tetris.fill = MAX_FILL;
		intCat(&buf1, "Number for fill out of bounds, use at most ",
			MAX_FILL);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tetris.fill);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

static void
drawUnit(TetrisWidget w, int pmid, int cid, int i, int j)
{
#ifdef IMAGE
	Boolean image = True;
#ifdef WINVER
	HBITMAP hOldBitmap, monoOldBitmap, hBitmap;
	HDC monoDC;
#endif
#endif

	if (i < 0 || i >= w->tetris.columns || j < 0 || j >= w->tetris.rows)
		return;
#ifdef IMAGE
	if (images[FALSE][pmid % BITMAPS] == None) {
		if (images[FALSE][BITMAPS - 1] != None)
			(void) printf("Bitmap unknown %d.\n", pmid);
		image = False;
	}
#ifdef WINVER
	if (image && (monoDC = CreateCompatibleDC(0)) == NULL) {
		(void) printf("Can not create mono DC.\n");
		image = False;
	}
	if (image && (w->core.memDC =
			CreateCompatibleDC(w->core.hDC)) == NULL) {
		(void) printf("Can not create compatible DC.\n");
		image = False;
	}
	if (image && (hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->tetris.boxSize, w->tetris.boxSize)) == NULL) {
		(void) printf("Can not create compatible bitmap.\n");
		image = False;
	}
#endif
	if (image) {
#ifdef WINVER
		SetTextColor(w->core.memDC, w->tetris.inverseGC[2]);
		SetBkColor(w->core.memDC, w->tetris.thingGC[cid]);
		FILLRECTANGLE(w, hBitmap, w->tetris.thingGC[cid],
			0, 0,
			w->tetris.boxSize, w->tetris.boxSize);
		monoOldBitmap = (HBITMAP) SelectObject(monoDC,
			images[FALSE][pmid % BITMAPS]);
		hOldBitmap = SelectObject(w->core.memDC, hBitmap);
		BitBlt(w->core.memDC,
			0, 0, w->tetris.boxSize, w->tetris.boxSize,
			monoDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(monoDC, monoOldBitmap);
		DeleteDC(monoDC);
		BitBlt(w->core.hDC,
			w->tetris.boxSize * i + w->tetris.offset.x,
			w->tetris.boxSize * j + w->tetris.offset.y,
			w->tetris.boxSize, w->tetris.boxSize,
			w->core.memDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(w->core.memDC, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(w->core.memDC);
#else
		(void) XPutImage(XtDisplay(w), XtWindow(w),
			w->tetris.thingGC[cid],
			images[FALSE][pmid % BITMAPS],
			0, 0,
			w->tetris.boxSize * i + w->tetris.offset.x,
			w->tetris.boxSize * j + w->tetris.offset.y,
			w->tetris.boxSize, w->tetris.boxSize);
#endif
	} else
#endif
	{
		FILLRECTANGLE(w, dr, w->tetris.thingGC[cid],
			w->tetris.boxSize * i + w->tetris.offset.x,
			w->tetris.boxSize * j + w->tetris.offset.y,
			w->tetris.boxSize, w->tetris.boxSize);
	}
}

static void
clearUnit(TetrisWidget w, int i, int j)
{
	if (i >= 0 && i < w->tetris.columns && j >= 0 && j < w->tetris.rows) {
		FILLRECTANGLE(w, dr, w->tetris.inverseGC[2],
			w->tetris.boxSize * i + w->tetris.offset.x,
			w->tetris.boxSize * j + w->tetris.offset.y,
			w->tetris.boxSize, w->tetris.boxSize);
	}
}

static void
xorUnit(TetrisWidget w, int i, int j)
{
#ifdef WINVER
	HDC memDC;
	HBITMAP hBitmap, hOldBitmap;

	if ((memDC = CreateCompatibleDC(w->core.hDC)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor memDC), exiting.");
	}
	if ((hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->tetris.boxSize,
			w->tetris.boxSize)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor hBitmap), exiting.");
	}
	hOldBitmap = SelectObject(memDC, hBitmap);
	BitBlt(memDC, 0, 0,
		w->tetris.boxSize, w->tetris.boxSize,
		w->core.hDC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		SRCCOPY);
	BitBlt(w->core.hDC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		w->tetris.boxSize, w->tetris.boxSize,
		memDC,
		0, 0,
		DSTINVERT);
	(void) SelectObject(memDC, hOldBitmap);
	DeleteObject(hBitmap);
	DeleteDC(memDC);
#else
	FILLRECTANGLE(w, dr, w->tetris.xorGC,
		w->tetris.boxSize * i + w->tetris.offset.x,
		w->tetris.boxSize * j + w->tetris.offset.y,
		w->tetris.boxSize, w->tetris.boxSize);
#endif
}

static void
moveUnit(TetrisWidget w, int ai, int aj, int bi, int bj)
{
#ifdef WINVER
	BitBlt(w->core.hDC,
		w->tetris.offset.x + bi * w->tetris.boxSize,
		w->tetris.offset.y + bj * w->tetris.boxSize,
		w->tetris.boxSize, w->tetris.boxSize,
		w->core.hDC,
		w->tetris.offset.x + ai * w->tetris.boxSize,
		w->tetris.offset.y + aj * w->tetris.boxSize,
		SRCCOPY);
#else
	VOID XSetGraphicsExposures(XtDisplay(w), w->tetris.inverseGC[2],
		False);
	VOID XCopyArea(XtDisplay(w), XtWindow(w), XtWindow(w),
		w->tetris.inverseGC[2],
		w->tetris.offset.x + ai * w->tetris.boxSize,
		w->tetris.offset.y + aj * w->tetris.boxSize,
		w->tetris.boxSize, w->tetris.boxSize,
		w->tetris.offset.x + bi * w->tetris.boxSize,
		w->tetris.offset.y + bj * w->tetris.boxSize);
#endif
}

static void
drawPartialField(TetrisWidget w, int start, int finish)
{
	int i, j, offset;

	for (j = start; j < finish; j++) {
		offset = j * w->tetris.columns;
		for (i = 0; i < w->tetris.columns; i++) {
			if (w->tetris.field[offset + i].pmid >= 0)
				drawUnit(w, w->tetris.field[offset + i].pmid,
					w->tetris.field[offset + i].cid,
					i, j);
		}
	}
}

static void
drawField(TetrisWidget w)
{
	drawPartialField(w, 0, w->tetris.rows);
}

static void
clearField(TetrisWidget w)
{
	int i, j;

	for (j = 0; j < w->tetris.rows; j++) {
		for (i = 0; i < w->tetris.columns; i++) {
			clearUnit(w, i, j);
		}
	}
}

static void
drawThing(TetrisWidget w)
{
	int i, j;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = w->tetris.curThing.xpos + i;
			int yj = w->tetris.curThing.ypos + j;

			if (yj >= 0 && tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape[j][i] != 0)

				drawUnit(w, polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape),
					w->tetris.curThing.size),
					w->tetris.curThing.colorNumber,
					xi, yj);
		}
}

static void
drawThingDiff(TetrisWidget w, Thing oldThing)
{
	int i, j;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = w->tetris.curThing.xpos + i;
			int yj = w->tetris.curThing.ypos + j;

			if (yj >= 0 && tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape[j][i] != 0)
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape),
					w->tetris.curThing.size),
					w->tetris.curThing.colorNumber,
					xi, yj);
		}
	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = oldThing.xpos + i;
			int yj = oldThing.ypos + j;
			int ox = xi - w->tetris.curThing.xpos;
			int oy = yj - w->tetris.curThing.ypos;

			if (yj >= 0 && tris[oldThing.units].polyomino
					[oldThing.polyNumber].shape[j][i] != 0 &&
					((ox < 0) || (ox >= w->tetris.curThing.size) ||
					(oy < 0) || (oy >= w->tetris.curThing.size) ||
					tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape[oy][ox] == 0))
				clearUnit(w, xi, yj);
		}
}

static void
draw3DFrame(const TetrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->tetris.inverseGC[3];
	else
		gc = w->tetris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->tetris.inverseGC[0];
	else
		gc = w->tetris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(TetrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tetris.inverseGC[2],
		0, 0, w->core.width, w->core.height);
}

static void
drawFrame(TetrisWidget w, Boolean focus)
{
	int offsetX = w->tetris.offset.x - 2;
	int offsetY = w->tetris.offset.y - 2;
	int lengthX = w->tetris.offset.x + w->tetris.viewLength + 2;
	int lengthY = w->tetris.offset.y + 2 * w->tetris.viewLength + 2;

	FILLRECTANGLE(w, dr, w->tetris.inverseGC[4],
		0, 0, w->core.width, offsetY);
	FILLRECTANGLE(w, dr, w->tetris.inverseGC[4],
		0, offsetY, offsetX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->tetris.inverseGC[4],
		lengthX, offsetY, w->core.width - lengthX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->tetris.inverseGC[4],
		0, lengthY, w->core.width, offsetY + 1);
	draw3DFrame(w, focus,
		offsetX, offsetY, lengthX - offsetX, lengthY - offsetY);
}

static void
drawNext(TetrisWidget w)
{
	setGameNext(w, ACTION_DRAW_NEXT,
		w->tetris.nextThing.randomNumber,
		w->tetris.nextThing.randomRotation,
		w->tetris.nextThing.randomReflection,
		w->tetris.units, w->tetris.diagonal, w->tetris.mixed,
		w->tetris.next.bonus);
}

static void
drawStatus(TetrisWidget w)
{
	setGame(w, ACTION_STATUS);
}

void
redoNext(TetrisWidget w)
{
	int diagonal = (w->tetris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->tetris.nextThing.units = w->tetris.next.units - MIN_UNITS +
		((w->tetris.nextThing.bonus) ? 1 : 0);
	next_start = tris[w->tetris.nextThing.units].mode
		[diagonal].number[(w->tetris.nextThing.bonus) ?
		NOMIX : ((w->tetris.next.mixed) ? 1 : 0)][turnStyle];
	if (!next_start) {
		DISPLAY_ERROR("Error, no polyominoes for this combination");
	}
	next_start = w->tetris.nextThing.randomNumber % next_start;
	w->tetris.nextThing.colorNumber = next_start;

	if (w->tetris.next.mixed && !w->tetris.nextThing.bonus) {
		w->tetris.nextThing.units = 0;
		while (next_start >= tris[w->tetris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->tetris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->tetris.nextThing.units++;
		}
	}
	w->tetris.nextThing.polyNumber =
		tris[w->tetris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->tetris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->tetris.nextThing.randomRotation; i++)
			w->tetris.nextThing.polyNumber =
				tris[w->tetris.nextThing.units].polyomino
				[w->tetris.nextThing.polyNumber].rotation;
	if (tris[w->tetris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->tetris.nextThing.randomReflection; i++)
			w->tetris.nextThing.polyNumber =
				tris[w->tetris.nextThing.units].polyomino
				[w->tetris.nextThing.polyNumber].reflection;
	w->tetris.nextThing.size = tris[w->tetris.nextThing.units].polyomino
		[w->tetris.nextThing.polyNumber].size;
	w->tetris.nextThing.xpos =
		NRAND(w->tetris.columns - w->tetris.nextThing.size + 1);
	/* nextThing.xpos = (w->tetris.columns - nextThing.size) / 2; */
	w->tetris.nextThing.ypos =
		-tris[w->tetris.nextThing.units].polyomino
		[w->tetris.nextThing.polyNumber].startHeight[(w->tetris.suddenAppear) ? JUMPIN : GRADUAL];
}

static void
initFill(TetrisWidget w)
{
	int i, j, col, offset, fixedUnits;

	if (w->tetris.fixedFill) {
		/* This algorithm randomly picks fixedUnits per row
		 * (4 for DEFAULT_COLUMNS, ie. 10/2-1)
		 * This is more in keeping with the original */
		fixedUnits = w->tetris.columns / 2 - 1;
		for (j = 0; j < w->tetris.rows; j++) {
			offset = j * w->tetris.columns;
			for (i = 0; i < w->tetris.columns; i++) {
				w->tetris.field[offset + i].pmid = -1;
				w->tetris.field[offset + i].cid = 0;
			}
			if (j >= w->tetris.rows - w->tetris.fill)
				for (i = 0; i < fixedUnits; i++) {
					do
						col = NRAND(w->tetris.columns);
					while (w->tetris.field[offset + col].pmid != -1);
					w->tetris.field[offset + col].pmid = BITMAPS - 1;
					w->tetris.field[offset + col].cid = NRAND(w->tetris.numberPolyominoes);
				}
		}
	} else {
		/* This algorithm randomly fills boxes to a certain level
		 * The problem with this is that it could potentially fill
		 * up a whole row */
		for (j = 0; j < w->tetris.rows; j++) {
			offset = j * w->tetris.columns;
			for (i = 0; i < w->tetris.columns; i++)
				if ((j >= w->tetris.rows - w->tetris.fill) &&
						(NRAND(2) == 0)) {
					w->tetris.field[offset + i].pmid = BITMAPS - 1;
					w->tetris.field[offset + i].cid = NRAND(w->tetris.numberPolyominoes);
				} else {
					w->tetris.field[offset + i].pmid = -1;
					w->tetris.field[offset + i].cid = 0;
				}
		}
	}
}

static void
fillLines(TetrisWidget w)
{
	int i, j;
	int diagonal = (w->tetris.current.diagonal) ? 1 : 0;
	int mixed = (w->tetris.current.mixed) ? 1 : 0;
	int turnStyle = tris[w->tetris.current.units -
		MIN_UNITS].mode[diagonal].turnStyle;

	FLUSH(w);
	for (j = 0; j <= w->tetris.rows / 2; j++) {
		for (i = 0; i < w->tetris.columns; i++) {
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, j);
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, w->tetris.rows - j - 1);
		}
		FLUSH(w);
		Sleep(100);
	}
	Sleep(1000);
}

static Boolean
overlapping(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++)
			if (tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyNumber].shape[j][i] != 0) {
				int xi = x + i;
				int yj = y + j;

				if ((yj >= w->tetris.rows) || (xi < 0) ||
						(xi >= w->tetris.columns))
					return True;
				if (w->tetris.suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
					if ((yj < 0) || (w->tetris.field[yj *
							w->tetris.columns + xi].pmid >= 0)) {
						return True;
					}
				} else { /* evolution */
/* This method one can turn polyomino to an area above of screen,
 * also part of the polyomino may not be visible initially */
					if ((yj >= 0) && (w->tetris.field[yj *
							w->tetris.columns + xi].pmid >= 0)) {
						return True;
					}
				}
			}
	return False;
}

static Boolean
atBottom(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = x + i;
			int yj = y + j;

			if ((yj >= -1) && tris[w->tetris.curThing.units].polyomino
				[w->tetris.curThing.polyNumber].shape[j][i] != 0)
				if ((yj >= w->tetris.rows - 1) ||
					(xi < 0) ||
					(xi >= w->tetris.columns) ||
					(w->tetris.field[(yj + 1) *
					w->tetris.columns + xi].pmid >= 0))
					return True;
		}
	return False;
}

static void
putBox(TetrisWidget w)
{
	int i, j;
	int x = w->tetris.curThing.xpos, y = w->tetris.curThing.ypos;

	for (i = 0; i < w->tetris.curThing.size; i++)
		for (j = 0; j < w->tetris.curThing.size; j++) {
			int xi = x + i;
			int yj = y + j;

			if ((yj >= 0) && tris[w->tetris.curThing.units].polyomino
				[w->tetris.curThing.polyNumber].shape[j][i]!= 0) {
				w->tetris.field[(yj) * w->tetris.columns + xi].pmid =
					(polyPixmap(i, j,
					(tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].shape),
					w->tetris.curThing.size)) % BITMAPS;
				w->tetris.field[(yj) * w->tetris.columns + xi].cid =
					w->tetris.curThing.colorNumber;
			}
		}
}

static int
checkLines(TetrisWidget w)
{
	int *lSet, nSet = 0;
	int i, j, k, y, offset;

	if (!(lSet = (int *) malloc(sizeof (int) * w->tetris.rows))) {
		DISPLAY_ERROR("Not enough memory (lSet), exiting.");
	}
	/* Find filled rows */
	for (j = 0; j < w->tetris.rows; j++) {
		lSet[j] = 0;
		offset = j * w->tetris.columns;
		for (i = 0; i < w->tetris.columns; i++)
			if (w->tetris.field[offset + i].pmid >= 0)
				lSet[j]++;
		if (lSet[j] == w->tetris.columns)
			nSet++;
	}

	if (nSet == 0) {
		free(lSet);
		return 0;
	}
	for (k = 0; k < 2 * NUM_FLASHES; k++) {
		for (j = 0; j < w->tetris.rows; j++) {
			if (lSet[j] == w->tetris.columns) {
				for (i = 0; i < w->tetris.columns; i++)
					xorUnit(w, i, j);
			}
		}
		FLUSH(w);
		Sleep(100);
	}
	/* Shrink */
	for (j = w->tetris.rows - 1; j >= 0; j--)
		if (lSet[j] == w->tetris.columns) {
			for (y = j; y > 0; y--)
				for (i = 0; i < w->tetris.columns; i++) {
					w->tetris.field[y * w->tetris.columns + i] =
						w->tetris.field[(y - 1) * w->tetris.columns + i];
					moveUnit(w, i, y - 1, i, y);
				}
			for (i = 0; i < w->tetris.columns; i++) {
				w->tetris.field[i].pmid = -1;
				clearUnit(w, i, 0);
			}
			for (i = j; i > 0; i--)
				lSet[i] = lSet[i - 1];
			lSet[0] = 0;
			if (j > 0)
				for (i = 0; i < w->tetris.columns; i++) {
					int tmp = w->tetris.field[j * w->tetris.columns + i].pmid;

					if ((tmp >= 0) && (tmp != CHECKDOWN(tmp))) {
						w->tetris.field[j * w->tetris.columns + i].pmid = CHECKDOWN(tmp);
						drawUnit(w, w->tetris.field[j * w->tetris.columns + i].pmid,
						w->tetris.field[j * w->tetris.columns + i].cid, i, j);
					}
				}
			j++;
			if (j < w->tetris.rows)
				for (i = 0; i < w->tetris.columns; i++) {
					int tmp = w->tetris.field[j * w->tetris.columns + i].pmid;

					if ((tmp >= 0) && (tmp != CHECKUP(tmp))) {
						w->tetris.field[j * w->tetris.columns + i].pmid = CHECKUP(tmp);
						drawUnit(w, w->tetris.field[j * w->tetris.columns + i].pmid,
						w->tetris.field[j * w->tetris.columns + i].cid, i, j);
					}
				}
			FLUSH(w);
#ifdef USE_SOUND
			if (w->tetris.sound) {
				playSound((char *) BUMPSOUND);
			}
#endif
		}
	free(lSet);
	return nSet;
}

static void
newThing(TetrisWidget w)
{
	w->tetris.curThing = w->tetris.nextThing;
	/*w->tetris.current = w->tetris.next;*/
	w->tetris.current.units = w->tetris.next.units;
	w->tetris.current.diagonal = w->tetris.next.diagonal;
	w->tetris.current.mixed = w->tetris.next.mixed;
	w->tetris.current.bonus = w->tetris.next.bonus;
	w->tetris.nextThing.randomNumber = LRAND();
	w->tetris.nextThing.randomRotation = NRAND(MAX_SIDES);
	w->tetris.nextThing.randomReflection = NRAND(2);
	w->tetris.nextThing.bonus = (w->tetris.next.bonus) ? 1 : 0;
	redoNext(w);
}

static void
clearThings(TetrisWidget w, Boolean draw)
{
	if (draw)
		clearField(w);
	initFill(w);
	if (draw)
		drawField(w);
	w->tetris.delay = delay[w->tetris.level - 1];
	w->tetris.score = w->tetris.fill * w->tetris.level * 10;
	w->tetris.started = True;
	w->tetris.paused = False;
	w->tetris.pauseStart = 0;
	w->tetris.completedRows = 0;
	w->tetris.newRows = 0;
	drawStatus(w);
}

static void
gameOver(TetrisWidget w, Boolean fast)
{
	if (!w->tetris.started)
		return;
	if (!fast)
		fillLines(w);
	setGame(w, ACTION_GAME_OVER);
	clearThings(w, True);
}

static void
tryMove(TetrisWidget w, move_t move)
{
	int diagonal = (w->tetris.current.diagonal) ? 1 : 0;
	Thing old;
	int i;

	old = w->tetris.curThing;

	switch (move) {
	case FALL:
		w->tetris.curThing.ypos++;
		break;

	case DROP:
		do {
			w->tetris.curThing.ypos++;
			if (w->tetris.dropPaused == 0)
				w->tetris.score += w->tetris.level +
					w->tetris.fill;
		} while (!overlapping(w));
		w->tetris.curThing.ypos--;
		break;

	case MOVE_LEFT:
		w->tetris.curThing.xpos--;
		break;

	case MOVE_RIGHT:
		w->tetris.curThing.xpos++;
		break;

	case ROTATE:
		if (tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle > NONE) {
			if (w->tetris.cw)
				for (i = 0; i < MAX_SIDES - 1; i++)
					w->tetris.curThing.polyNumber =
						tris[w->tetris.curThing.units].polyomino
						[w->tetris.curThing.polyNumber].rotation;
			else
				w->tetris.curThing.polyNumber =
					tris[w->tetris.curThing.units].polyomino
					[w->tetris.curThing.polyNumber].rotation;
			w->tetris.curThing.xpos = old.xpos;
			w->tetris.curThing.ypos = old.ypos;
		}
		break;

	case REFLECT: /* reflect on y axis */
		if (tris[w->tetris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle == ALL) {
		w->tetris.curThing.polyNumber =
			tris[w->tetris.curThing.units].polyomino
			[w->tetris.curThing.polyNumber].reflection;
		w->tetris.curThing.xpos = old.xpos;
		w->tetris.curThing.ypos = old.ypos;
		}
		break;
	}
	if (overlapping(w)) {
		w->tetris.curThing = old;
	} else {
		drawThingDiff(w, old);
	}
}

static Boolean
moveOne(TetrisWidget w, move_t move)
{
	int lines;

	if ((move == DROP) || ((move == FALL) && atBottom(w))) {
		tryMove(w, move);
		putBox(w);
		lines = checkLines(w);
		if (w->tetris.dropPaused != 0) {
			w->tetris.dropPaused = 0;
		} else {
			w->tetris.score += (w->tetris.fill + w->tetris.level +
				lines + ((w->tetris.suddenAppear) ? 1 : 0)) *
				((w->tetris.current.bonus) ? 1 : 3) *
				((w->tetris.showNext) ? 1 : 3) /
				(1 + w->tetris.current.mixed);
			w->tetris.score += ((w->tetris.showNext) ? 3 : 5) *
				w->tetris.level * w->tetris.level * lines * lines *
				((w->tetris.current.bonus) ? 3 : 5) *
				w->tetris.startLevel *
				(w->tetris.current.units + ((w->tetris.current.diagonal) ? 1 : 0)) *
				(w->tetris.current.units + ((w->tetris.current.diagonal) ? 1 : 0)) /
				(16 * (1 + ((w->tetris.current.mixed) ? 1 : 0)));
		}
		w->tetris.completedRows += lines;
		w->tetris.newRows += lines;
		w->tetris.next.bonus = False;
		if ((w->tetris.newRows > THRESHOLD(w->tetris.level)) &&
				(w->tetris.level < MAX_LEVELS)) {
			w->tetris.level++;
			w->tetris.newRows = 0;
			if (LEGAL(w->tetris.next.units,
					w->tetris.next.diagonal,
					w->tetris.next.mixed,
					w->tetris.bonus))
				w->tetris.next.bonus = True;
		}
		drawStatus(w);
		newThing(w);
		drawThing(w);
		drawNext(w);
		FLUSH(w);
		if (overlapping(w)) {
			gameOver(w, False);
		}
		return True;
	} else {
		if (w->tetris.dropPaused == 1)
			w->tetris.dropPaused = 2;
		tryMove(w, move);
		if ((w->tetris.newRows > THRESHOLD(w->tetris.level)) &&
				(w->tetris.level < MAX_LEVELS)) {
			w->tetris.next.bonus = False;
			w->tetris.level++;
			w->tetris.newRows = 0;
			if (LEGAL(w->tetris.next.units,
					w->tetris.next.diagonal,
					w->tetris.next.mixed,
					w->tetris.bonus))
				w->tetris.next.bonus = True;
			drawStatus(w);
			drawNext(w);
		}
		return False;
	}
}

#ifndef WINVER
static
#endif
void
falldownGame(TetrisWidget w)
{
	if (!w->tetris.paused) {
		w->tetris.gotNew = moveOne(w, FALL) || w->tetris.gotNew;
	} else {
		if (w->tetris.pauseStart >= 32) {
			clearThings(w, True);
		}
		if (!w->tetris.started)
			w->tetris.pauseStart++;
	}
}

#ifndef WINVER
static
#endif
void
moveGameInput(TetrisWidget w, int direction)
{
	if (!w->tetris.paused)
		w->tetris.gotNew = moveOne(w, (move_t) direction) ||
			w->tetris.gotNew;
}

static void
resetThings(TetrisWidget w)
{
	if (w->tetris.field)
		free(w->tetris.field);
	w->tetris.rows = 2 * w->tetris.columns;
	if (!(w->tetris.field = (Field *) malloc(
			w->tetris.rows * w->tetris.columns * sizeof(Field)))) {
		DISPLAY_ERROR("Not enough memory (field), exiting.");
	}
	w->tetris.started = False;
}

static void
resizeThings(TetrisWidget w)
{
#ifdef IMAGE
#ifndef WINVER
	Display *display = XtDisplay(w);
	Window window = XtWindow(w);
	XWindowAttributes xgwa;

	(void) XGetWindowAttributes(display, window, &xgwa);
#endif
	free_images(FALSE);
	if (w->tetris.boxSize > 4)
		(void) create_images(
#ifndef WINVER
			display, xgwa.visual,
#endif
			(w->tetris.plainPiece) ? 0 : 1, FALSE,
			w->tetris.boxSize);
#endif
}

#define BRIGHT_FACTOR 0.8
#define DARK_FACTOR 0.75

#ifdef WINVER
#define MAX_INTENSITY 0xFF

int
brighter(const int light)
{
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);
	int temp = light;

	if (temp < i)
		temp = i;
	return MIN(temp / BRIGHT_FACTOR, MAX_INTENSITY);
}

int
darker(const int light)
{
	return (int) (light * DARK_FACTOR);
}

static void
setValuesGame(TetrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->tetris.level = GetPrivateProfileInt(SECTION,
		"level", DEFAULT_LEVELS, INIFILE);
	w->tetris.units = GetPrivateProfileInt(SECTION,
		"units", DEFAULT_UNITS, INIFILE);
	w->tetris.columns = GetPrivateProfileInt(SECTION,
		"columns", DEFAULT_COLUMNS, INIFILE);
	w->tetris.fill = GetPrivateProfileInt(SECTION,
		"fill", 0, INIFILE);
	w->tetris.fixedFill = (BOOL) GetPrivateProfileInt(SECTION,
		"fixedFill", TRUE, INIFILE);
	w->tetris.suddenAppear = (BOOL) GetPrivateProfileInt(SECTION,
		"suddenAppear", FALSE, INIFILE);
	w->tetris.diagonal = (BOOL) GetPrivateProfileInt(SECTION,
		"diagonal", DEFAULT_DIAGONAL, INIFILE);
	w->tetris.mixed = (BOOL) GetPrivateProfileInt(SECTION,
		"mixed", DEFAULT_MIXED, INIFILE);
	w->tetris.bonus = (BOOL) GetPrivateProfileInt(SECTION,
		"bonus", TRUE, INIFILE);
	w->tetris.cw = (BOOL) GetPrivateProfileInt(SECTION,
		"cw", TRUE, INIFILE);
	w->tetris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->tetris.showNext = (BOOL) GetPrivateProfileInt(SECTION,
		"showNext", FALSE, INIFILE);
	w->tetris.sound = (BOOL) GetPrivateProfileInt(SECTION,
		"sound", FALSE, INIFILE);
	(void) GetPrivateProfileString(SECTION,
		"bumpSound", BUMPSOUND, szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tetris.bumpSound, szBuf, 80);
	w->tetris.mono = (BOOL) GetPrivateProfileInt(SECTION,
		"mono", DEFAULT_MONO, INIFILE);
	w->tetris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION,
		"background", "255 255 255", szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tetris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->tetris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->tetris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->tetris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->tetris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION,
			buf, polyominoColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->tetris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
	}
	(void) GetPrivateProfileString(SECTION,
		"userName", "Guest", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tetris.userName, szBuf, 80);
	(void) GetPrivateProfileString(SECTION,
		"scoreFile", "", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->tetris.scoreFile, szBuf, 80);
}

void
destroyGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
#define MAX_INTENSITY 0xFFFF

Pixel
brighter(TetrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / BRIGHT_FACTOR, MAX_INTENSITY);
	color.green = (unsigned short) MIN(color.green / BRIGHT_FACTOR, MAX_INTENSITY);
	color.blue = (unsigned short) MIN(color.blue / BRIGHT_FACTOR, MAX_INTENSITY);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Tetris: Could not allocate brighter color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

Pixel
darker(TetrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	color.red = (unsigned short) (color.red * DARK_FACTOR);
	color.green = (unsigned short) (color.green * DARK_FACTOR);
	color.blue = (unsigned short) (color.blue * DARK_FACTOR);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Tetris: Could not allocate darker color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

static void
getColor(TetrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->tetris.reverse) {
		values.background = w->tetris.foreground;
	} else {
		values.background = darker(w, w->tetris.background);
	}
	if (!w->tetris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tetris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tetris.reverse) {
			values.foreground = w->tetris.background;
		} else {
			values.foreground = w->tetris.foreground;
		}
	}
	if (w->tetris.thingGC[thing])
		XtReleaseGC((Widget) w, w->tetris.thingGC[thing]);
	w->tetris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
}

static void
setAllColors(TetrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyomino;

	valueMask = GCForeground | GCBackground;

	if (w->tetris.reverse) {
		values.background = w->tetris.background;
		values.foreground = w->tetris.foreground;
	} else {
		values.foreground = w->tetris.background;
		values.background = w->tetris.foreground;
	}
	if (w->tetris.inverseGC[1])
		XtReleaseGC((Widget) w, w->tetris.inverseGC[1]);
	w->tetris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tetris.mono) {
		values.foreground = brighter(w, (w->tetris.reverse) ?
			w->tetris.foreground : w->tetris.background);
	}
	if (w->tetris.inverseGC[0])
		XtReleaseGC((Widget) w, w->tetris.inverseGC[0]);
	w->tetris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tetris.mono) {
		values.foreground = darker(w, (w->tetris.reverse) ?
			w->tetris.foreground : w->tetris.background);
	}
	if (w->tetris.inverseGC[2])
		XtReleaseGC((Widget) w, w->tetris.inverseGC[2]);
	w->tetris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tetris.mono) {
		values.foreground = darker(w, darker(w, (w->tetris.reverse) ?
			w->tetris.foreground : w->tetris.background));
	}
	if (w->tetris.inverseGC[3])
		XtReleaseGC((Widget) w, w->tetris.inverseGC[3]);
	w->tetris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tetris.mono) {
		values.foreground = darker(w, darker(w, darker(w,
			(w->tetris.reverse) ? w->tetris.foreground :
			w->tetris.background)));
	}
	if (w->tetris.inverseGC[4])
		XtReleaseGC((Widget) w, w->tetris.inverseGC[4]);
	w->tetris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyomino = 0; polyomino < MAX_START_POLYOMINOES; polyomino++)
		getColor(w, polyomino);
	valueMask = GCForeground | GCBackground | GCFunction;
	values.foreground = 0xFFFFFF;
	values.background = 0;
	values.function = GXxor;
	if (w->tetris.xorGC)
		XtReleaseGC((Widget) w, w->tetris.xorGC);
	w->tetris.xorGC = XtGetGC((Widget) w, valueMask, &values);
}

static Boolean
setValuesGame(Widget current, Widget request, Widget renew)
{
	TetrisWidget c = (TetrisWidget) current, w = (TetrisWidget) renew;
	Boolean redraw = False, setColors = False;

	if (w->tetris.act) {
		falldownGame(w);
		w->tetris.act = False;
	}
	if (w->tetris.level != c->tetris.level)
		w->tetris.delay = delay[w->tetris.level - 1];
	checkThings(w);
	if (w->tetris.reverse != c->tetris.reverse ||
			w->tetris.mono != c->tetris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->tetris.background != c->tetris.background ||
			w->tetris.foreground != c->tetris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->tetris.plainPiece != c->tetris.plainPiece) {
		redraw = True;
	} else if (w->tetris.suddenAppear != c->tetris.suddenAppear) {
                redoNext(w);
	}
	if (w->tetris.columns != c->tetris.columns) {
		w->tetris.paused = True;
		w->tetris.started = False;
		eraseFrame(w);
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, False);
		drawFrame(w, w->tetris.focus);
		redraw = True;
	} else if (w->tetris.fill != c->tetris.fill) {
		clearThings(w, True);
		redraw = True;
	} else if (w->tetris.units != c->tetris.units) {
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.diagonal != c->tetris.diagonal) {
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.mixed != c->tetris.mixed) {
		w->tetris.next.mixed = w->tetris.mixed;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->tetris.bonus != c->tetris.bonus) {
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
	}
	if (w->tetris.menu != -1) {
		switch (w->tetris.menu) {
		case 1:
			w->tetris.paused = True;
			break;
		case 2:
			w->tetris.menu = -1;
			clearThings(w, True);
			redraw = True;
			break;
		case 4:
			w->tetris.menu = -1;
			/*ConfirmThings(w);*/
			redraw = True;
			break;
		default:
			w->tetris.menu = -1;
			break;
		}
	}
	return (redraw);
}

static void
quitGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	gameOver(w, True);
	XtCloseDisplay(display);
	exit(0);
}

static void
destroyGame(Widget old)
{
	TetrisWidget w = (TetrisWidget) old;
	int i;

#if defined( USE_SOUND ) && defined( USE_ESOUND )
	(void) shutdown_sound();
#endif

	for (i = 0; i < MAX_START_POLYOMINOES; i++)
		XtReleaseGC(old, w->tetris.thingGC[i]);
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->tetris.inverseGC[i]);
	XtReleaseGC(old, w->tetris.xorGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->tetris.select);
}
#endif

#ifndef WINVER
static
#endif
void
resizeGame(TetrisWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->tetris.boxSize = MIN((w->core.height - 4) / 2, w->core.width - 4) /
		w->tetris.columns;
	w->tetris.viewLength = w->tetris.boxSize * w->tetris.columns;
	w->tetris.offset.x = ((int) w->core.width + 1 -
		w->tetris.viewLength) / 2;
	w->tetris.offset.y = ((int) w->core.height + 1 -
		2 * w->tetris.viewLength) / 2;
}

#ifndef WINVER
static
#endif
void
sizeGame(TetrisWidget w)
{
	resetThings(w);
	resizeGame(w);
}

#ifndef WINVER
static
#endif
void
initializeGame(
#ifdef WINVER
TetrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifdef IMAGE
	int bitmap;
#endif
#ifndef WINVER
	int i;
	TetrisWidget w = (TetrisWidget) renew;
#endif

#ifdef WINVER
	setValuesGame(w);
#else
	w->tetris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->tetris.mono);
	for (i = 0; i < MAX_START_POLYOMINOES; i++)
		w->tetris.thingGC[i] = NULL;
	for (i = 0; i < BG_SHADES; i++)
		w->tetris.inverseGC[i] = NULL;
	w->tetris.xorGC = NULL;
#endif
#ifdef IMAGE
	for (bitmap = 0; bitmap < BITMAPS; bitmap++)
		images[FALSE][bitmap] = None;
#endif
	w->tetris.field = NULL;
	w->tetris.started = False;
	w->tetris.quit = False;
	w->tetris.completedRows = 0;
	w->tetris.newRows = 0;
	w->tetris.fall = False;
	w->tetris.focus = False;
	w->tetris.next.units = w->tetris.units;
	w->tetris.next.diagonal = w->tetris.diagonal;
	w->tetris.next.mixed = w->tetris.mixed;
	w->tetris.next.bonus = False;
	w->tetris.startLevel = w->tetris.level;
	w->tetris.nextThing.units = 0;
	w->tetris.nextThing.polyNumber = 0;
	w->tetris.nextThing.xpos = 0;
	w->tetris.nextThing.ypos = 0;
	w->tetris.nextThing.size = 0;
	w->tetris.nextThing.colorNumber = 0;
	w->tetris.nextThing.bonus = 0;
	w->tetris.nextThing.randomRotation = 0;
	w->tetris.nextThing.randomReflection = 0;
	w->tetris.nextThing.randomNumber = 0;
	w->tetris.gotNew = False;
	w->tetris.dropPaused = 0;
	checkThings(w);
	sizeGame(w);
#ifdef WINVER
	brush = CreateSolidBrush(w->tetris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	diagonal = (w->tetris.next.diagonal) ? 1 : 0;
	mixed = (w->tetris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "Tetris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->tetris.numberPolyominoes = tris[w->tetris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->tetris.numberPolyominoes <= 0 ||
			w->tetris.numberPolyominoes > MAX_START_POLYOMINOES) {
		(void) fprintf(stderr, "Tetris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tnumberPolyominoes = %d.\n",
			w->tetris.numberPolyominoes);
		exit(1);
	}
	clearThings(w, False);
	newThing(w);
	newThing(w);
#ifdef USE_SOUND
#ifdef USE_NAS
	dsp = XtDisplay(w);
#endif
#ifdef USE_ESOUND
	(void) init_sound();
#endif
#endif
}

#ifndef WINVER
static
#endif
void
exposeGame(
#ifdef WINVER
TetrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	TetrisWidget w = (TetrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
	resizeThings(w);
	drawFrame(w, w->tetris.focus);
	drawField(w);
	drawThing(w);
	drawNext(w);
}

#ifndef WINVER
static
#endif
void
hideGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tetris.sound)
		w->tetris.sound = False;
	w->tetris.paused = True;
	w->tetris.dropPaused = 2;
	setGame(w, ACTION_HIDE);
}

#ifndef WINVER
static
#endif
void
clearGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	clearThings(w, True);
}

#ifndef WINVER
static
#endif
void
enterGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tetris.focus = True;
	w->tetris.paused = False;
	drawFrame(w, w->tetris.focus);
}

#ifndef WINVER
static
#endif
void
leaveGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	/* if (w->tetris.firstFall) */
	w->tetris.focus = False;
	w->tetris.paused = True;
	if (w->tetris.dropPaused == 0)
		w->tetris.dropPaused = 1;
	else
		w->tetris.dropPaused = 2;
	w->tetris.dropPaused = True;
	drawFrame(w, w->tetris.focus);
}

#ifndef WINVER
static void
mouseGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	if (w->tetris.dropPaused == 1) /* clicking causes enter/exit events */
		w->tetris.dropPaused = 0;
}

static void
reflectGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, REFLECT);
}

static void
rotateGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, ROTATE);
}

static void
moveGameLeft(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_LEFT);
}

static void
moveGameRight(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_RIGHT);
}

static void
dropGame(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, DROP);
}

#endif

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameLevel(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.level < MAX_LEVELS) {
		w->tetris.level++;
		w->tetris.delay = delay[w->tetris.level - 1];
		w->tetris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#ifdef WINVER
		return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameLevel(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
#ifdef CHEAT
	if (!w->tetris.paused && w->tetris.level > MIN_LEVELS) {
		w->tetris.level--;
		w->tetris.delay = delay[w->tetris.level - 1];
		w->tetris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#endif
#ifdef WINVER
		return False;
#endif
}

#ifndef WINVER
static
#endif
void
pauseGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->tetris.sound)
		w->tetris.sound = False;
	w->tetris.paused = !w->tetris.paused;
	w->tetris.dropPaused = 2;
#if 0
	if (w->tetris.paused) {
		banner(MSG_PAUSED);
		(void) XIconifyWindow(display, mainWin, screen_num);
	} else {
		drawNext(w);
	}
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameUnits(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units + 1,
			w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, ACTION_INC_UNITS);
#ifdef WINVER
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameUnits(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units - 1,
			w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, ACTION_DEC_UNITS);
#ifdef WINVER
		w->tetris.next.units = w->tetris.units;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameColumns(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.columns < MAX_COLUMNS) {
#ifdef WINVER
		w->tetris.paused = True;
		w->tetris.started = False;
		eraseFrame(w);
		w->tetris.columns++;
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->tetris.focus);
		return True;
#else
		setGame(w, ACTION_INC_COLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameColumns(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && w->tetris.columns > MIN_COLUMNS) {
#ifdef WINVER
		w->tetris.paused = True;
		w->tetris.started = False;
		eraseFrame(w);
		w->tetris.columns--;
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->tetris.focus);
		return True;
#else
		setGame(w, ACTION_DEC_COLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
fixedfillGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_FIXED_FILL);
}

#ifndef WINVER
static
#endif
void
suddenAppearGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SUDDEN_APPEAR);
#ifdef WINVER
	redoNext(w);
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
diagonalGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units,
			!w->tetris.next.diagonal, w->tetris.next.mixed,
			False)) {
		setGame(w, ACTION_DIAGONAL);
#ifdef WINVER
		w->tetris.next.diagonal = w->tetris.diagonal;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
mixedGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused && LEGAL(w->tetris.next.units,
			w->tetris.next.diagonal, !w->tetris.next.mixed,
			False)) {
		setGame(w, ACTION_MIXED);
#ifdef WINVER
		w->tetris.next.mixed = w->tetris.mixed;
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
bonusGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
		setGame(w, ACTION_BONUS);
#ifdef WINVER
		w->tetris.next.bonus = False;
		w->tetris.newRows = 0;
		redoNext(w);
		drawNext(w);
#endif
}

#ifndef WINVER
static
#endif
void
cwGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_CW);
}

#ifndef WINVER
static
#endif
void
plainPieceGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_PLAIN_PIECE);
#ifdef WINVER
	resizeThings(w);
	drawField(w);
#endif
}

#ifndef WINVER
static
#endif
void
showNextGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tetris.paused) {
		setGame(w, ACTION_SHOW_NEXT);
	}
}

#ifndef WINVER
static
#endif
void
soundGame(TetrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SOUND);
}

#ifndef WINVER
static
#endif
void
fillGame(TetrisWidget w, int n)
{
	setGame(w, n + ACTION_FILL0);
#ifdef WINVER
	clearThings(w, True);
#endif
}

#ifndef WINVER
static void
fill0Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 0);
}

static void
fill1Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 1);
}

static void
fill2Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 2);
}

static void
fill3Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 3);
}

static void
fill4Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 4);
}

static void
fill5Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 5);
}

static void
fill6Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 6);
}

static void
fill7Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 7);
}

static void
fill8Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 8);
}

static void
fill9Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 9);
}

static void
fill10Game(TetrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 10);
}
#endif
